#!/usr/bin/perl -w
# Author: Andrew Sitnikov
# Copyright: 2002 Andrew Sitnikov
# $Id: $

use strict;
use lib '/usr/lib/mysqlstat';

$|=1;

use CGI qw(:standard);
use CGI::Carp qw(fatalsToBrowser);
use HTML::Template;
use POSIX qw(strftime);
use Digest::MD5  qw(md5_hex);

use mysqlstat;

###############################################################################
# CONFIG BEGIN
###############################################################################
use constant HOME_DIR 		 => '/usr/share/mysqlstat';
use constant WORK_DIR 		 => '/var/lib/mysqlstat';

use constant SKIN_BASE_DIR	 => HOME_DIR.'/skins';
use constant CACHE_DIR 		 => '/var/lib/mysqlstat/cache';
###############################################################################
# CONFIG END
###############################################################################

###############################################################################
# SETTINGS BEGIN
###############################################################################
my @MODE_LIST = (
	{ name => 'Queries',   	 		value => 'query',		current => 0 , version => 32200 },
	{ name => 'In/Out',   	 		value => 'in_out',		current => 0 , version => 32300 },
	{ name => 'Query type',			value => 'com',			current => 0 , version => 32347 },
	{ name => 'Tmp tables usage',	value => 'tmp',			current => 0 , version => 32300 },
	{ name => 'Key reads',     		value => 'key_read',	current => 0 , version => 32300 },
	{ name => 'Key writes',    		value => 'key_write',	current => 0 , version => 32300 },
	{ name => 'Query cache',   		value => 'query_cache',	current => 0 , version => 40000 },	
	{ name => 'Status',    			value => 'status',		current => 0 , version => 32200 },
);

my @PERIOD_LIST = (
	{ name => 'Last 6h',    value => 'h', current => 0 },
	{ name => 'Last 2d' ,   value => 'd', current => 0 },
	{ name => 'Last week',  value => 'w', current => 0 },
	{ name => 'Last month', value => 'm', current => 0 },
	{ name => 'Last year',  value => 'y', current => 0 },
);

my @SKIN_LIST = getSkinList(SKIN_BASE_DIR);
my @TARGET_LIST = ();

###############################################################################
# SETTINGS END
###############################################################################
my @current_target = ();

my $target	= param('t') || '';
my $mode	= param('m') || 'query';
my $period	= param('p') || 'd';
my $graph	= param('g') || 0;
my $skin	= param('skin') || '';

$skin =~ s/[^\w]//g;

if ($skin &&  -d SKIN_BASE_DIR.'/'.$skin){
	$ENV{HTML_TEMPLATE_ROOT} = SKIN_BASE_DIR.'/'.$skin;
}else{
	$ENV{HTML_TEMPLATE_ROOT} = SKIN_BASE_DIR.'/default';
}

my $rdata = getRetentionData(WORK_DIR,1);

print_error("Can't read retension data") unless($rdata);

$target = '' unless(exists($rdata->{servers}->{$target}));

if ($target){
  my $t = $rdata->{servers}->{$target};
  my $tver = ($t->{status} && exists($t->{status}->{Version}))?$t->{status}->{Version}:0;
  
  if ($tver =~ /(\d+)\.(\d+)\.(\d+)/){
	$tver = $1*10000+$2*100+$3;
	@MODE_LIST = grep { $tver>=$_->{version} } @MODE_LIST;
  }
}

my $found = 0;
foreach(@PERIOD_LIST){
	last if ($_->{value} eq $period && ($found = $_->{current} = 1));
}

unless($found){
  $period = $PERIOD_LIST[0]->{value};
  $PERIOD_LIST[0]->{current} = 1;
}

$found = 0;
foreach(@MODE_LIST){
	last if ($_->{value} eq $mode && ($found = $_->{current} = 1));
}

unless($found){
  $mode = $MODE_LIST[0]->{value};
  $MODE_LIST[0]->{current} = 1;
}

foreach(@SKIN_LIST){
	last if ($_->{value} eq $skin && ($_->{current} = 1));
}

if (!$graph){
	@TARGET_LIST = getTargetList($rdata->{servers});
}


if (!$target){
	print_global_status();
}elsif (!$graph){
	print_detail_status($target,$mode,$period);
}else{
  print_img($target,$mode,$period);
}

###############################################################################
###############################################################################
sub print_error{
	my $template = HTML::Template->new(filename => 'error.tmpl', die_on_bad_params => 0, global_vars => 1);
	$template->param(MESSAGE => join("\n",@_));

	print header,$template->output;
	exit;
}

###############################################################################
###############################################################################
sub print_global_status{
	my $template = HTML::Template->new(filename => 'global.tmpl', die_on_bad_params => 0, global_vars => 1);

	$template->param(
		CSKIN			=> $skin,
		SKIN_LIST	 	=> @SKIN_LIST ? \@SKIN_LIST : 0,

		TARGET_LIST   	=> \@TARGET_LIST,

		GLOBAL_Started 		=> strftime("%Y:%m: %d %H:%M", localtime($rdata->{Started})),
		GLOBAL_LastUpdated 	=> strftime("%Y:%m: %d %H:%M", localtime($rdata->{LastUpdated})),

		GLOBAL_Total_Questions => formatQuery($rdata->{Total_Questions}),
		GLOBAL_Total_In        => formatTrafic($rdata->{Total_In}),
		GLOBAL_Total_Out       => formatTrafic($rdata->{Total_Out}),
	);

	print header, $template->output;
}

###############################################################################
###############################################################################
sub print_detail_status {
	my $target = shift;
	my $mode = shift;
	my $period = shift;

	my $template = HTML::Template->new(filename => 'detail.tmpl', die_on_bad_params => 0, global_vars => 1);

	if ($mode eq 'status'){
		my @TARGET_STATUS = getTargetStatus($target);

		$template->param(
			TARGET_STATUS    => \@TARGET_STATUS,
		);
	}

	$template->param(
		CTARGET => $target,
		CMODE 	=> $mode,
		CPERIOD => $period,
		CSKIN	=> $skin,

		CURRENT_TARGET	=> \@current_target,
		TARGET_LIST		=> \@TARGET_LIST,

		MODE_LIST		=> \@MODE_LIST,
		PERIOD_LIST		=> \@PERIOD_LIST,
		SKIN_LIST		=> \@SKIN_LIST
	);

	print header,$template->output;
}

###############################################################################
###############################################################################
sub print_img {
	my $target = shift;
	my $mode = shift;
	my $period = shift;

	my $rrd_file = WORK_DIR.'/mysqlstat-'.$target.'.rrd';
	my $server = $rdata->{servers}->{$target};

	my (@options,@DEF,@PLOT,@LEGEND);

	push(@options,qw(-w 420 -h 150 --interlaced));

	my $title = $server->{name};

	if ($server->{status}){
		$title .= ', Version: '.$server->{status}->{Version};
		$title .= ', Uptime: '.formatUptime($server->{status}->{Uptime});
	}

	push(@options,'-t',$title); #FIXME Prepare name for shell chars

	if ($period eq 'h'){
		push(@options,qw(--start -6h));
	}elsif ($period eq 'w'){
		push(@options,qw(--start -1w));
	}elsif($period eq 'm'){
		push(@options,qw(--start -1m));
	}elsif($period eq 'y'){
		push(@options,qw(--start -1y));
	}

	if ($mode eq 'in_out'){
		push(@DEF,'DEF:I='.$rrd_file.':Bytes_received:AVERAGE');
		push(@DEF,'DEF:O='.$rrd_file.':Bytes_sent:AVERAGE');

		push(@PLOT,
			'AREA:O#0000FF:Output',
			'GPRINT:O:LAST:(current=%.0lf',
			'GPRINT:O:AVERAGE:ave=%.0lf',
			'GPRINT:O:MAX:max=%.0lf)',
			'COMMENT:\n',
		);

		push(@PLOT,
			'AREA:I#FFFF00:Input',
			'GPRINT:I:LAST:(current=%.0lf',
			'GPRINT:I:AVERAGE:ave=%.0lf',
			'GPRINT:I:MAX:max=%.0lf)',
			'COMMENT:\n',
		);

	}elsif($mode eq 'com'){
#		push(@options,'-v','Q/sec');

		push(@DEF,'DEF:S='.$rrd_file.':Com_select:AVERAGE');
		push(@DEF,'DEF:I='.$rrd_file.':Com_insert:AVERAGE');
		push(@DEF,'DEF:D='.$rrd_file.':Com_delete:AVERAGE');
		push(@DEF,'DEF:U='.$rrd_file.':Com_update:AVERAGE');
		push(@DEF,'DEF:C='.$rrd_file.':Com_change_db:AVERAGE');

		push(@PLOT,
			'AREA:S#0000FF:Select',
			'GPRINT:S:LAST:(current=%.0lf',
			'GPRINT:S:AVERAGE:ave=%.0lf',
			'GPRINT:S:MAX:max=%.0lf Q/sec)',
			'COMMENT:\n',
		);

		push(@PLOT,
			'STACK:C#FF00FF:Change DB',
			'GPRINT:C:LAST:(current=%.0lf',
			'GPRINT:C:AVERAGE:ave=%.0lf',
			'GPRINT:C:MAX:max=%.0lf Q/sec)',
			'COMMENT:\n',
		);

		push(@PLOT,
			'STACK:U#FFFF00:Update',
			'GPRINT:U:LAST:(current=%.0lf',
			'GPRINT:U:AVERAGE:ave=%.0lf',
			'GPRINT:U:MAX:max=%.0lf Q/sec)',
			'COMMENT:\n',
		);

		push(@PLOT,
			'STACK:I#00FF00:Insert',
			'GPRINT:I:LAST:(current=%.0lf',
			'GPRINT:I:AVERAGE:ave=%.0lf',
			'GPRINT:I:MAX:max=%.0lf Q/sec)',
			'COMMENT:\n',
		);

		push(@PLOT,
			'STACK:D#FF0000:Delete',
			'GPRINT:D:LAST:(current=%.0lf',
			'GPRINT:D:AVERAGE:ave=%.0lf',
			'GPRINT:D:MAX:max=%.0lf Q/sec)',
			'COMMENT:\n',
		);

	}elsif($mode eq 'tmp'){
#		push(@options,'-v','Q/sec');

		push(@DEF,'DEF:T='.$rrd_file.':Created_tmp_tables:AVERAGE');
		push(@DEF,'DEF:D='.$rrd_file.':Created_tmp_disk_ta:AVERAGE');

		push(@PLOT,
			'AREA:T#0000FF:Created tmp tables',
			'GPRINT:T:LAST:(current=%.0lf',
			'GPRINT:T:AVERAGE:ave=%.0lf',
			'GPRINT:T:MAX:max=%.0lf)',
			'COMMENT:\n',
		);

		push(@PLOT,
			'AREA:D#FFFF00:Created tmp disk tables',
			'GPRINT:D:LAST:(current=%.0lf',
			'GPRINT:D:AVERAGE:ave=%.0lf',
			'GPRINT:D:MAX:max=%.0lf)',
			'COMMENT:\n',
		);
	}elsif($mode eq 'key_read'){
#		push(@options,'-v','Q/sec');

		push(@DEF,'DEF:T='.$rrd_file.':Key_read_requests:AVERAGE');
		push(@DEF,'DEF:R='.$rrd_file.':Key_reads:AVERAGE');

		push(@PLOT,
			'AREA:T#0000FF:Key read requests',
			'GPRINT:T:LAST:(current=%.0lf',
			'GPRINT:T:AVERAGE:ave=%.0lf',
			'GPRINT:T:MAX:max=%.0lf)',
			'COMMENT:\n',
		);

		push(@PLOT,
			'AREA:R#FFFF00:Key_reads',
			'GPRINT:R:LAST:(current=%.0lf',
			'GPRINT:R:AVERAGE:ave=%.0lf',
			'GPRINT:R:MAX:max=%.0lf)',
			'COMMENT:\n',
		);

	}elsif($mode eq 'key_write'){
#		push(@options,'-v','Q/sec');

		push(@DEF,'DEF:T='.$rrd_file.':Key_write_requests:AVERAGE');
		push(@DEF,'DEF:W='.$rrd_file.':Key_writes:AVERAGE');

		push(@PLOT,
			'AREA:T#0000FF:Key write requests',
			'GPRINT:T:LAST:(current=%.0lf',
			'GPRINT:T:AVERAGE:ave=%.0lf',
			'GPRINT:T:MAX:max=%.0lf)',
			'COMMENT:\n',
		);

		push(@PLOT,
			'AREA:W#FFFF00:Key_writes',
			'GPRINT:W:LAST:(current=%.0lf',
			'GPRINT:W:AVERAGE:ave=%.0lf',
			'GPRINT:W:MAX:max=%.0lf)',
			'COMMENT:\n',
		);
	}elsif($mode eq 'query_cache'){
		push(@options,'-v','Q/sec');

		push(@DEF,'DEF:S='.$rrd_file.':Com_select:AVERAGE');
		push(@DEF,'DEF:I='.$rrd_file.':Qcache_inserts:AVERAGE');
		push(@DEF,'DEF:H='.$rrd_file.':Qcache_hits:AVERAGE');
		push(@DEF,'DEF:N='.$rrd_file.':Qcache_not_cached:AVERAGE');

		push(@PLOT,
			'AREA:S#FFFF00:Command SELECT',
			'GPRINT:S:LAST:(current=%.0lf',
			'GPRINT:S:AVERAGE:ave=%.0lf',
			'GPRINT:S:MAX:max=%.0lf)',
			'COMMENT:\n',
		);

		push(@PLOT,
			'AREA:I#0000FF:Qcache Inserts',
			'GPRINT:I:LAST:(current=%.0lf',
			'GPRINT:I:AVERAGE:ave=%.0lf',
			'GPRINT:I:MAX:max=%.0lf)',
			'COMMENT:\n',
		);

		push(@PLOT,
			'AREA:H#00FF00:Qcache Hits',
			'GPRINT:H:LAST:(current=%.0lf',
			'GPRINT:H:AVERAGE:ave=%.0lf',
			'GPRINT:H:MAX:max=%.0lf)',
			'COMMENT:\n',
		);

		push(@PLOT,
			'AREA:N#FF0000:Qcache not cached',
			'GPRINT:N:LAST:(current=%.0lf',
			'GPRINT:N:AVERAGE:ave=%.0lf',
			'GPRINT:N:MAX:max=%.0lf)',
			'COMMENT:\n',
		);


	}else{
		push(@options,'-v','Q/sec');

		push(@DEF,'DEF:Q='.$rrd_file.':Questions:AVERAGE');
		push(@DEF,'DEF:S='.$rrd_file.':Slow_queries:AVERAGE');
		push(@DEF,'DEF:C='.$rrd_file.':Connections:AVERAGE');
		push(@DEF,'DEF:M='.$rrd_file.':Max_used_connection:AVERAGE');

		push(@PLOT,
			'AREA:Q#0000FF:Queries',
			'GPRINT:Q:LAST:(current=%.0lf',
			'GPRINT:Q:AVERAGE:ave=%.0lf',
			'GPRINT:Q:MAX:max=%.0lf Q/sec)',
			'COMMENT:\n',
		);

		push(@PLOT,
			'AREA:C#FFFF00:Connections',
			'GPRINT:C:LAST:(current=%.0lf',
			'GPRINT:C:AVERAGE:ave=%.0lf',
			'GPRINT:C:MAX:max=%.0lf C/sec)',
			'COMMENT:\n',
		);

		push(@PLOT,
			'LINE1:S#FF0000:Slow queries',
			'GPRINT:S:LAST:(current=%.0lf',
			'GPRINT:S:AVERAGE:ave=%.0lf',
			'GPRINT:S:MAX:max=%.0lf Q/sec)',
			'COMMENT:\n',
		);

#		push(@PLOT,
#			'LINE2:M#FF0000:Max used connections',
#			'GPRINT:M:LAST:(current=%.0lf',
#			'GPRINT:M:AVERAGE:ave=%.0lf',
#			'GPRINT:M:MAX:max=%.0lf Q/sec)',
#			'COMMENT:\n',
#		);
	}

	#print header;
	print header('image/gif');

	my $cache_file_name = 'mysqlstat-'.md5_hex(join('',@options, @DEF, @PLOT, @LEGEND));
	my $out_file = '-';
	my $cached = 0;

	if ( -d CACHE_DIR && -w CACHE_DIR){
		$out_file = CACHE_DIR.'/'.$cache_file_name;
		$cached = 1;
		push(@options,'--lazy');
	}elsif( -w '/dev/fd/1' ){
		$out_file = '/dev/fd/1';
	}

  RRDs::graph($out_file, @options, @DEF, @PLOT, @LEGEND);
  my $ERR = RRDs::error;
  warn "RRDs: $ERR" if $ERR;

  if ($cached){
  	local $/ = undef;

  	open(IMG,"< $out_file") or die("Can't open cached image");
  	print <IMG>;
  	close(IMG);
  }
}

sub getSkinList{
	my $skin_dir = shift;
	my @skin_list;

	return () unless(opendir(SKIN_DIR,$skin_dir));

	foreach(readdir(SKIN_DIR)){
		next if (/^\./ || ! -d $skin_dir.'/'.$_);

		my ($value,$name,$author) = ($_,$_);

		if ( -f SKIN_BASE_DIR.'/'.$_.'/skin.info'){
			if (open(FIN,"< ".SKIN_BASE_DIR.'/'.$_.'/skin.info')){
			  while(<FIN>){
					chomp;
					if(/Name:\s*(.+)$/){
						$name = $1;
						next;
					}elsif(/Author:\s*(.+)$/){
						$author = $1;
						next;
					}
				}
				close(FIN);
			}
		}

		$name = $_ unless($name);
		$author = '' unless($author);

		push(@skin_list,{ name=>$name, value=>$value, author =>$author });
	}

	closedir SKIN_DIR;

	return sort { lc($$a{name}) cmp lc($$b{name}) } @skin_list;
}

sub getTargetList{
  my $server_list = shift;
	my @targel_list;

	foreach(keys %{ $server_list }){
	  my $server = $server_list->{$_};

	  my $data = {
			'target' 				 => $_,
			'name'   				 => $server->{name},
			'disabled'	 		 => $server->{disabled},

			'StartTime'      => strftime("%Y:%m: %d %H:%M", localtime($server->{StartTime})),
			'LastUpdate'     => strftime("%Y:%m: %d %H:%M", localtime($server->{LastUpdate})),
			'Current'				 => ($target eq $_),
		};

		if ($server->{status}){
			$data->{Version} = $server->{status}->{Version};
			$data->{Uptime} = formatUptime($server->{status}->{Uptime});
			$data->{Bytes_received} = formatTrafic($server->{status}->{Bytes_received});
			$data->{Bytes_sent} = formatTrafic($server->{status}->{Bytes_sent});
	  }else{
			$data->{Version} = 'UNK';
			$data->{Uptime} = 'UNK';
			$data->{Bytes_received} = 'UNK';
			$data->{Bytes_sent} = 'UNK';
		}

		push(@targel_list,$data);

		if (($target eq $_) && $server->{status}){
			push(@current_target,{ 'name' => $_, 'value' => $server->{status}->{$_} }) foreach(sort keys %{ $server->{status} });
		}
	}

	return sort { lc($a->{name}) cmp lc($b->{name}) }	@targel_list;;
}

sub getTargetStatus
{
	my $target = shift;

	my @s = ();

	return @s unless($rdata->{servers}->{$target}->{status});

	foreach(sort keys %{ $rdata->{servers}->{$target}->{status} }){
		push(@s, { name => $_, value => $rdata->{servers}->{$target}->{status}->{$_} } );
	}

	return @s;
}

# vim:ts=2:sw=2
